import sys

def _trans_VerbForm(item):
  key, value = item
  _dict = {
    "Inf": "原形",
    "Fin": None,
    #"Ger": "",
  }
  result = _dict[value] if value in _dict.keys() else "{}={}".format(key, value)
  return  result 

def _trans_Mood(item):
  key, value = item
  _dict = {
    "Ind": "陈述式",
    "Imp": "命令式",
    "Cnd": "条件式",
    "Sub": "虚拟式",
  }
  result = _dict[value] if value in _dict.keys() else "{}={}".format(key, value)
  return  result 

def _trans_Tense(item):
  key, value = item
  _dict = {
    "Pres": "一般现在时",
    "Past": "简单过去时",
    "Fut": "将来时",
    "Imp": "未完成时",
  }
  result = _dict[value] if value in _dict.keys() else "{}={}".format(key, value)
  return  result 


def _trans_Number(item):
  key, value = item
  _dict = {
    "Sing": "单数",
    "Plur": "复数",
  }
  result = _dict[value] if value in _dict.keys() else "{}={}".format(key, value)
  return  result 

def _trans_Person(item):
  key, value = item
  _dict = {
    "1": "第一人称",
    "2": "第二人称",
    "3": "第三人称",
  }
  result = _dict[value] if value in _dict.keys() else "{}={}".format(key, value)
  return  result 

def _trans_PronType(item):
  key, value = item
  _dict = {
    "Prs": "代词",
  }
  result = _dict[value] if value in _dict.keys() else "{}={}".format(key, value)
  return  result 

def _trans_Reflex(item):
  key, value = item
  _dict = {
    "Yes": "自复",
  }
  result = _dict[value] if value in _dict.keys() else "{}={}".format(key, value)
  return  result 

def trans_morph(morph):
  if "Tense=Past" in morph and "VerbForm=Part" in morph:
    return "过去分词"
  morph_dict = morph.to_dict()
  ordered = ["VerbForm", "Mood", "Tense", "Person", "Number", "Reflex", "PronType", "Case"]
  translated = []
  for key in ordered:
    if key in morph_dict.keys():
      current_pkg = sys.modules[__name__] 
      fn_name = "_trans_{}".format(key)
      value = morph_dict[key]
      if hasattr(current_pkg, fn_name):
        fn = getattr(current_pkg, fn_name)
        translated.append(fn((key,value)))
      else:
        translated.append("{}={}".format(key, value))
  return "|".join(list(filter(None, translated)))



from spacy.language import Language 
from spacy.tokens import Doc, Token
from structure_detective.lib import form_children_info #, get_children_except_xs

vp1s = ["come on", "think of", "look for", "meet with", "call upon", "object to", "consist of", "make by", "make for", "look after", "agree to", "look into", "abide by", "abound in"]
vp2s = ["look down on", "look forward to", "look out for", "look out on", "carry on with", "catch up with", "come down on", "come out against", "date back to", "do away with", "face up to", "feel up to"]

def _is_verb_like(e):
  return e.pos_ in ["VERB", "AUX"]

def _has_no_subject(e):
  children_deps = [x.dep_ for x in e.children]
  return all([subj not in children_deps for subj in ["nsubj", "nsubjpass"]])

def _get_root(doc: Doc):
  roots = [r for r in doc if r.dep_=="ROOT"]
  if len(roots)>=1:
    root = roots[0]
  else:
    root = None
  return root

def _extend_root_by_vp_2(doc: Doc, root: Token):
  roots, childrens = [root], []
  if root.i+2>=len(doc):
    childrens = [[t for t in root.children]]
    return roots, childrens
  nt1, nt2 = doc[root.i+1], doc[root.i+2]
  phrase = ""
  validation = all([nt1.dep_ in ["prep", "advmod", "prt"], nt1.head == root, nt2.head in [root, nt1]])
  if validation:
    children = list(nt2.children)
    if len(children)==0 or (len(children) == 1 and children[0].dep_ == "pobj"):
      phrase = " ".join([root.lemma_, nt1.text, nt2.text])
    if phrase in vp2s:
      roots.extend([nt1, nt2])
      childrens = [[t for t in root.children if t not in [nt1, nt2]],[t for t in nt2.children]]
    else:
      childrens = [[t for t in root.children]]
  return roots, childrens

def _extend_root_by_vp_1(doc: Doc, root: Token):
  roots, childrens = [root], []
  if root.i+1>=len(doc):
    childrens = [[t for t in root.children]]
    return roots, childrens
  nt = doc[root.i+1]
  phrase = ""
  if nt.dep_ == "prep" and nt.head == root:
    children = list(nt.children)
    if len(children)==0 or (len(children) == 1 and children[0].dep_ == "pobj"):
      phrase = " ".join([root.lemma_, nt.text])
  if phrase in vp1s:
    roots.append(nt)
    childrens = [[t for t in root.children if t!=nt],[t for t in nt.children]] 
  else:
    childrens = [[t for t in root.children]]
  return roots, childrens

def _extend_root_by_half_have_to(doc: Doc, root: Token):
  roots, childrens = [root], []
  if root.i+1>=len(doc):
    childrens = [[t for t in root.children]]
    return roots, childrens
  nt = doc[root.i+1]
  phrase = " ".join([root.lemma_, nt.text])
  validation = all([phrase=="have to", nt.dep_=="aux", nt.head.dep_=="xcomp", nt.head.head==root])
  if validation:
    xcomp_token = nt.head
    roots.append(xcomp_token)
    childrens = [[t for t in root.children if t!=xcomp_token],[t for t in xcomp_token.children if t!=nt]] 
  else:
    childrens = [[t for t in root.children]]
  return roots, childrens
   

def _extend_root_by_half_have_got_to(doc: Doc, root: Token):
  roots, childrens = [root], []
  if root.i+1>=len(doc):
    childrens = [[t for t in root.children]]
    return roots, childrens
  nt = doc[root.i+1]
  phrase = " ".join([root.text, nt.text])
  aux_token_lemmas = [t.lemma_ for t in root.children if t.dep_=="aux"]
  validation = all([("have" in aux_token_lemmas or "'ve" in aux_token_lemmas) and phrase=="got to", nt.dep_=="aux", nt.head.dep_=="xcomp", nt.head.head==root])
  if validation:
    xcomp_token = nt.head
    roots.append(xcomp_token)
    childrens = [[t for t in root.children if t!=xcomp_token],[t for t in xcomp_token.children if t!=nt], [nt]] 
  else:
    childrens = [[t for t in root.children]]
  return roots, childrens

def _extend_root_by_half_be_going_to(doc: Doc, root: Token):
  roots, childrens = [root], []
  if root.i+1>=len(doc):
    childrens = [[t for t in root.children]]
    return roots, childrens
  nt = doc[root.i+1]
  phrase = " ".join([root.text, nt.text])
  aux_token_lemmas = [t.lemma_ for t in root.children if t.dep_=="aux"]
  validation = all([("be" in aux_token_lemmas) and phrase=="going to", nt.dep_=="aux", nt.head.dep_=="xcomp", nt.head.head==root])
  if validation:
    xcomp_token = nt.head
    roots.append(xcomp_token)
    childrens = [[t for t in root.children if t!=xcomp_token],[t for t in xcomp_token.children if t!=nt], [nt]] 
  else:
    childrens = [[t for t in root.children]]
  return roots, childrens

def _handle_vp_root(doc:Doc, roots: list):
  store = []
  _r_range = range(len(roots))
  for index in _r_range:
    if index < len(roots)-1:
      assert(roots[index].i+1 == roots[index+1].i)
  # append ROOT
  root = roots[0] 
  start, end = root.i, roots[-1].i+1
  _info = {
    "start": start,
    "end": end,
    "text": doc[start:end].text,
    "element": root.i, 
    "is_root": True,
    "semantic_dep": "ROOT"
  }
  store.append(_info)
  return store

def _handle_half_have_to_root(doc:Doc, roots: list):
  store = []
  # append ROOT
  for root in roots:
    start = root.i
    end = root.i+2 if root.dep_== "ROOT" and len(roots)==2 else root.i+1
    _info = {
      "start": start,
      "end": end,
      "text": doc[start:end].text,
      "element": root.i, 
      "is_root": root.dep_=="ROOT",
      "semantic_dep": "ROOT"
    }
    store.append(_info)
  return store

def _handle_half_have_got_to_root(doc:Doc, roots: list):
  store = []
  # append ROOT
  for root in roots:
    start = root.i
    end = root.i+1
    _info = {
      "start": start,
      "end": end,
      "text": doc[start:end].text,
      "element": root.i, 
      "is_root": root.dep_=="ROOT",
      "semantic_dep": "ROOT"
    }
    store.append(_info)
  return store

def _handle_half_be_going_to_root(doc:Doc, roots: list):
  store = []
  # append ROOT
  for root in roots:
    start = root.i
    end = root.i+1
    _info = {
      "start": start,
      "end": end,
      "text": doc[start:end].text,
      "element": root.i, 
      "is_root": root.dep_=="ROOT",
      "semantic_dep": "ROOT"
    }
    store.append(_info)
  return store


def _extend_root(doc: Doc, root: Token):
  # make rules
  if root.lemma_ == "have":
    roots, childrens = _extend_root_by_half_have_to(doc, root)
    store = _handle_half_have_to_root(doc, roots)
  elif root.text == "got":
    roots, childrens = _extend_root_by_half_have_got_to(doc, root)
    store = _handle_half_have_got_to_root(doc, roots)
  elif root.text == "going":
    roots, childrens = _extend_root_by_half_be_going_to(doc, root)
    store = _handle_half_be_going_to_root(doc, roots)
  else:
    roots, childrens = _extend_root_by_vp_2(doc, root)
    if len(roots) < 3:
      roots, childrens = _extend_root_by_vp_1(doc, root)
    store = _handle_vp_root(doc, roots)

  for children in childrens: 
    store.extend(form_children_info(doc, children))
  return store 

def get_tree(doc: Doc, nlp: Language):
  root = _get_root(doc)
  if root == None:
    return []

  store = []
  if _is_verb_like(root):
    store = _extend_root(doc, root) 
  ordered_store = sorted(store, key=lambda d: d["start"]) 
  return ordered_store 




from spacy.language import Language 
from spacy.tokens import Doc, Token

def has_dep(token:Token, dep:str):
  token_children_deps = [t.dep_ for t in token.children]
  return dep in token_children_deps


def get_children_except_xs(token:Token, xs:list):
  return [t for t in token.children if t.dep_ not in xs] 
  
def get_subtree_except_xs(token:Token, xs:list):
  xs_subtrees = [list(x.subtree) for x in token.children if x.dep_ in xs]

  excluded_subtree = sum(xs_subtrees, [])
  included_subtree = [s for s in token.subtree if s not in excluded_subtree]
  return included_subtree
 
def is_top(t:Token, ts:list):
  return all([t.is_ancestor(x) for x in ts if t!=x])

def form_children_info(doc:Doc, children:list):
  store = []
  #print(doc.text)
  for dc in children:
    dc_ranges = [e.i for e in dc.subtree]
    dc_ranges.sort()
    dc_children_ids = [x.i for x in dc.children]
    #print(dc_children_ids)

    length = len(dc_ranges)
    current = 0
    subtrees = [[dc_ranges[0]]]
    for i in range(length-1):
      a, b = dc_ranges[i], dc_ranges[i+1] 
      if a == b-1:
        subtrees[current].append(b) 
      else:
        current = current + 1
        subtrees.append([b])
    #print(subtrees)
    #print("="*10)
    for t in subtrees: 
      tokens = [doc[x] for x in t]
      
      heads = [h for h in tokens if is_top(h, tokens)]
      assert(len(heads)==1)
      head = heads[0]
        
      start, end = min(t), max(t)+1
      _info = {
        "start": start,
        "end": end,
        "text": doc[start:end].text,
        "element": head.i, 
        "is_root": False,
        "semantic_dep": head.dep_
      }
      store.append(_info)
  return store





from spacy.language import Language 
from spacy.tokens import Doc
from structure_detective.lib import form_children_info, get_children_except_xs

def _is_verb_like(e):
  return e.pos_ in ["VERB", "AUX"]

def _has_no_subject(e):
  children_deps = [x.dep_ for x in e.children]
  return all([subj not in children_deps for subj in ["sb", "sbp", "pm"]])

def _get_root(doc: Doc):
  print(doc.text)
  roots = [r for r in doc if r.dep_=="ROOT"]
  if len(roots)>=1:
    root = roots[0]
  else:
    root = None
  return root

def _handle_backward_ocs(ocs:list, doc:Doc):
  store = []
  if len(ocs) == 0:
    return store
  else:
    for oc in ocs:
      regular_oc_children = get_children_except_xs(oc, ["oc"])
      store.extend(form_children_info(doc, regular_oc_children))

      start, end = oc.i, oc.i+1
      _info = {
        "start": start,
        "end": end,
        "text": doc[start:end].text,
        "element": oc.i,
        "is_root": False,
      }
      store.append(_info)
 
      oc_children = [occ for occ in oc.children if occ.dep_=="oc"]
      store.extend(_handle_backward_ocs(oc_children, doc))
    return store

def _handle_oc(doc: Doc, nlp: Language):
  store = []
  # append ROOT
  root = _get_root(doc)
  start, end = root.i, root.i+1
  _info = {
    "start": start,
    "end": end,
    "text": doc[start:end].text,
    "element": root.i, 
    "is_root": True,
  }
  store.append(_info)

  # No oc
  ocs =  [t for t in root.children if t.dep_ == "oc"]
  if len(ocs) == 0:
    return store

  oc = ocs[0]
  
  no_need_to_dig = not _is_verb_like(oc) or "sb" in [t.dep_ for t in oc.children] or "ep" in [t.dep_ for t in oc.children]
  if no_need_to_dig:
    # oc IS NOT verb like
    store.extend(form_children_info(doc, [oc]))
    return store
  else:
    # oc IS verb like 
    regular_oc_children = get_children_except_xs(oc, ["pm", "oc"])
    store.extend(form_children_info(doc, regular_oc_children))

    # handle oc node with/without zu
    zu_children = [t for t in oc.children if t.dep_ in ["pm"] and t.text == "zu" and t.i == oc.i-1]
    if len(zu_children) == 1:
      start, end = oc.i-1, oc.i+1 
    else:
      start, end = oc.i, oc.i+1
    _info = {
      "start": start,
      "end": end,
      "text": doc[start:end].text,
      "element": oc.i,
      "is_root": False,
    }
    store.append(_info)

    # handle forward oc
    oc_forward_children = [t for t in oc.children if t.dep_ == "oc" and t.i>oc.i]
    store.extend(form_children_info(doc, oc_forward_children))

    # handle backward oc
    oc_backward_children = [t for t in oc.children if t.dep_ == "oc" and t.i<oc.i]
    store.extend(_handle_backward_ocs(oc_backward_children, doc))
    return store    

def get_tree(doc: Doc, nlp: Language):
  root = _get_root(doc)
  if root == None:
    return []

  if _is_verb_like(root):
    xs = ["oc"]
    children = [r for r in root.children if r.dep_ not in xs]
    store = form_children_info(doc, children)
  
    oc_store = _handle_oc(doc, nlp)
    store.extend(oc_store)
  else:
    store = []

  ordered_store = sorted(store, key=lambda d: d["start"]) 
  return ordered_store 




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['structure_detective',
 'structure_detective.de',
 'structure_detective.en',
 'structure_detective.es']

package_data = \
{'': ['*']}

install_requires = \
['spacy>=3.2.0,<4.0.0']

setup_kwargs = {
    'name': 'structure-detective',
    'version': '0.1.3',
    'description': 'Parse sentence',
    'long_description': None,
    'author': 'Phoenix.Grey',
    'author_email': 'phoenix.grey0108@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
